define(
	[
		'backbone'
		, 'mustache'
		, 'MegaPixImage'
		, 'js/edit/controller/EditController'
		, 'js/edit/model/EditModel'
		, 'js/edit/util/EditUtil'
		, 'js/edit/util/GraphicsUtil'
		, 'js/edit/util/ThreeUtil'
		, 'js/edit/util/DataUtil'
		, 'js/edit/util/ViewModeUtil'
		, 'js/index/util/LocalUtil'
		, 'js/common/model/Point'
		, 'js/common/model/Cuboid'
		//, 'js/common/model/Cuboid3d'
		, 'text!js/common/template/nav.html'
		, 'text!js/edit/template/body.html'
	],
	function(
		Backbone
		, Mustache
		, MegaPixImage
		, editController  // オブジェクト
		, editModel       // オブジェクト
		, editUtil        // オブジェクト
		, graphicsUtil    // オブジェクト
		, threeUtil       // オブジェクト
		, dataUtil        // オブジェクト
		, viewModeUtil    // オブジェクト
		, localUtil       // オブジェクト
		, Point           // クラス
		, Cuboid          // クラス
		//, Cuboid3d        // クラス
		, nav             // HTML
		, body            // HTML
	) {
		var EditView = Backbone.View.extend({
			//=================================================================
			// イベント
			events: {
				// ファイル選択
				  'change #imageFile1': 'changeImageFile1'
				, 'change #imageFile2': 'changeImageFile2'

				// 枠をリセットボタン
				, 'click #resetCuboid1': 'clickResetCuboid1'
				, 'click #resetCuboid2': 'clickResetCuboid2'

				// 画像情報ボタン
				, 'click #imageInfo1': 'clickImageInfo1'
				, 'click #imageInfo2': 'clickImageInfo2'

				// キャンバスイベント canvas1
				, 'touchstart  #canvas1': 'touchstartCanvas1'  // スクリーンに触れた
				, 'touchend    #canvas1': 'touchendCanvas1'    // スクリーンから指が離れた
				, 'touchmove   #canvas1': 'touchmoveCanvas1'   // スクリーンに触れている指をスライドさせた
				, 'touchcancel #canvas1': 'touchcancelCanvas1' // システムによってタッチがキャンセルされた
				, 'mousedown   #canvas1': 'mousedownCanvas1'
				, 'mouseup     #canvas1': 'mouseupCanvas1'
				, 'mousemove   #canvas1': 'mousemoveCanvas1'

				// キャンバスイベント canvas2
				, 'touchstart  #canvas2': 'touchstartCanvas2'  // スクリーンに触れた
				, 'touchend    #canvas2': 'touchendCanvas2'    // スクリーンから指が離れた
				, 'touchmove   #canvas2': 'touchmoveCanvas2'   // スクリーンに触れている指をスライドさせた
				, 'touchcancel #canvas2': 'touchcancelCanvas2' // システムによってタッチがキャンセルされた
				, 'mousedown   #canvas2': 'mousedownCanvas2'
				, 'mouseup     #canvas2': 'mouseupCanvas2'
				, 'mousemove   #canvas2': 'mousemoveCanvas2'

				// 3D データを生成
				, 'click  #create3d': 'clickCreate3d'
				
				// 3D データをコントロール
				, 'touchstart    #preview3d': 'touchstartPreview3d'    // スクリーンに触れた
				, 'touchend      #preview3d': 'touchendPreview3d'      // スクリーンから指が離れた
				, 'touchmove     #preview3d': 'touchmovePreview3d'     // スクリーンに触れている指をスライドさせた
				, 'touchcancel   #preview3d': 'touchcancelPreview3d'   // システムによってタッチがキャンセルされた
				, 'mousedown     #preview3d': 'mousedownPreview3d'
				, 'mouseup       #preview3d': 'mouseupPreview3d'
				, 'mousemove     #preview3d': 'mousemovePreview3d'
				
				// 3D データをコントロール (ジェスチャー)
				, 'gesturestart #preview3d':  'gesturestartPreview3d'
				, 'gesturechange #preview3d': 'gesturechangePreview3d'
				, 'gestureend #preview3d':    'gestureendPreview3d'
				
				// 3D 自動回転
				, 'change #autoRotation' : 'changeAutoRotation'
				
				// 3D サイズ調整
				, 'change #cuboidWidth' : 'changeCuboidWidth'
				, 'change #cuboidHeight': 'changeCuboidHeight'
				, 'change #cuboidDepth' : 'changeCuboidDepth'
				
				// 保存・アップロード
				, 'click #saveLocal': 'clickSaveLocal'
				, 'click #upload': 'clickUpload'
				
			},

			//=================================================================
			// 初期化処理
			initialize: function() {
				this.$el.html(Mustache.render(body, { nav: nav }));
				window.onresize = this.resizeWindow;  // ウィンドウサイズ変更イベント登録

				//-----------------------------------------
				// File API 使用可否確認
				if(window.File) {
				} else {
					alert("お使いのブラウザは機能が不足しているためご利用出来ません。");
				}
				
				//-----------------------------------------
				// ハッシュに応じて動きを変える
				var hash = location.hash;
				if(hash.length > 1) {
					var paramList = hash.split(':');
					if(paramList.length === 2) {
						var cuboid = null;
						var id;

						//-----------------------------------------
						// 一定の条件を満たすのであれば、即時閲覧専用モードに移行
						editModel.mode = 'view';  // モード指定
						viewModeUtil.setup();     // ページセットアップ
						
						//-----------------------------------------
						if(paramList[0] === '#local') {
							// ローカルストレージからデータを取得 (全部取得する必要はまったくない)
							id = parseInt(paramList[1]);
							var err = false;
							var cubox = localUtil.getCuboxData();
							if(cubox && cubox.cuboids && cubox.cuboids.length && cubox.cuboids.length > 0) {
								if(id >= cubox.cuboids.length) {
									err = true;
								} else {
									cuboid = cubox.cuboids[id];  // CuboidJSON
									viewModeUtil.setupPreview3d(cuboid, this);
								}
							} else {
								err = true;
							}
							if(err) {
								alert('指定されたIDは無効です。');
							}
							
						} else if(paramList[0] === '#remote') {
							// メッセージ表示
							$('#viewModeCuboidInfo').html('loading... <img src="img/loader.gif">');
							
							// サーバからデータを取得 (async)
							var This = this;
							id = paramList[1];
							$.ajax({
								url: 'cuboid/' + id + '.json',
								dataType: 'json',
								async: false
							}).done(function(data, status, jqxhr) {
								viewModeUtil.setupPreview3d(data, This);
							}).fail(function(jqxhr, status, error) {
								$('#viewModeCuboidInfo').text('データのロードに失敗しました。(load failed)');
							});
							
						}
					}
				}

				//-----------------------------------------
				// メニューアクティブ化
				if(editModel.mode === 'edit') {
					$('#navEdit').attr('class', 'active');
				}

			},

			//=================================================================
			// ウィンドウサイズ変更
			resizeWindow: function() {
				setTimeout(function() {
					editController.resizeCanvas(1);
					editController.resizeCanvas(2);
					clearTimeout();
				}, 50);
			},

			//=================================================================
			// ファイル選択
			changeImageFile1: function() { this.changeImageFile(1); },
			changeImageFile2: function() { this.changeImageFile(2); },
			changeImageFile: function(id) {
				// File API 使用可否確認
				if(window.File) {
					var reader = new FileReader();
					var file = $('#imageFile' + id).get(0).files[0];
					if(file && file.type.match('image.*')) {
						// ファイルのロードが完了した時に画像を表示する処理を登録
						reader.onload = (function(theFile, theId) {
							return function(e) {
								var binaryFile = new BinaryFile(e.target.result);

								// 画像データ保持
								editModel.exif[id - 1] = EXIF.readFromBinaryFile(binaryFile);
								editModel.prettyExif[id - 1] = editController.createPrettyExif(id);

								// 画像を表示
								$('#image' + theId).get(0).src = 'data:image/jpeg;base64,' + binaryFile.toBase64(); //e.target.result;

								// ボタン表示
								$('#resetCuboid' + id).css('display', '');
								$('#imageInfo' + id).css('display', '');

								// 次の入力項目を有効化
								if(id === 1) {  // step.2 を有効にする
									$('#imageFile2').removeAttr("disabled");
								} else {  // step.3 を有効にする
									$('#create3d').removeAttr("disabled");
								}

								// キャンバスサイズを変更 (少し待って実施)
								setTimeout(function(id) {
									// キャンバスサイズ変更
									editController.resizeCanvas(id);
									
									// オブジェクト生成
									editModel.cuboids[id-1] = new Cuboid();
									editModel.cuboids[id-1].init(id);
									
									// キャンバス再描画
									editController.redraw(id);

									// タイマー削除
									clearTimeout();
								}, 500, theId);
							};
						})(file, id);
						
						// バイナリファイルロード
						reader.readAsBinaryString(file);

						// iOS6 バグ回避のため別口でロード
						editModel.megaPixImage[id - 1] = new MegaPixImage(file);
					}
				}
			},

			//=================================================================
			// 枠をリセット
			clickResetCuboid1: function(e) { this.clickResetCuboid(e, 1); },
			clickResetCuboid2: function(e) { this.clickResetCuboid(e, 2); },
			clickResetCuboid: function(e, id) {
				var $canvas = $('#canvas' + id);
				if($canvas.length) {
					// 枠をリセット
					editModel.cuboids[id-1].init(id);

					// キャンバス再描画
					editController.redraw(id);
				}
			},

			//=================================================================
			// 画像情報
			clickImageInfo1: function(e) { this.clickImageInfo(e, 1); },
			clickImageInfo2: function(e) { this.clickImageInfo(e, 2); },
			clickImageInfo: function(e, id) {
				alert(editModel.prettyExif[id - 1]);
			},


			//=================================================================
			// 頂点のタッチ判定
			isTouchedPoint: function(x, y, ox, oy) {
				if( x >= ox - editModel.touchRange &&
					x <= ox + editModel.touchRange &&
					y >= oy - editModel.touchRange &&
					y <= oy + editModel.touchRange
				) {
					return true;
				}
				return false;
			},

			//=================================================================
			// キャンバスタッチ開始
			// 最も近いものだけを動かす。同じ指で動かせる頂点は1個まで。
			touchstartCanvas1: function(e) { this.touchstartCanvas(e, 1); },
			touchstartCanvas2: function(e) { this.touchstartCanvas(e, 2); },
			touchstartCanvas: function(e, id) {
				var touch, x, y, ox, oy, ax, ay;
				var touches = e.originalEvent.touches;  // 現在のタッチ情報
				var $canvas = $('#canvas' + id);
				var cuboid = editModel.cuboids[id-1];
				
				// フラグ初期化
				var exist = false;
				
				// 実行
				if(touches.length) {
					for(var j=0; j<touches.length; j++) {
						touch = touches[j];
						ox = touch.clientX - $canvas.offset().left + $(document).scrollLeft();
						oy = touch.clientY - $canvas.offset().top + $(document).scrollTop();
						for(var i=0; i<cuboid.points.length; i++) {
							x = cuboid.points[i].x;
							y = cuboid.points[i].y;
							if( this.isTouchedPoint(x, y, ox, oy) ) {
								// 自分よりもタッチ座標に近い、「moving 状態でない他の頂点」がない場合のみ動かす
								exist = false;
								for(var k=0; k<cuboid.points.length; k++) {
									if(k != i) {
										ax = cuboid.points[k].x;
										ay = cuboid.points[k].y;
										if( this.isTouchedPoint(ax, ay, ox, oy) ) {  // これを入れてはダメ → && cuboid.points[k].moving === false
											// i より近い頂点があるのなら、この頂点は動かすのにふさわしくない
											if(
												Math.sqrt(Math.pow(ox - x, 2) + Math.pow(oy - y, 2)) >    // i
												Math.sqrt(Math.pow(ox - ax, 2) + Math.pow(oy - ay, 2))    // k
											) {
												exist = true;
												break;
											}
										}
									}
								}
								if(!exist) {
									cuboid.points[i].moving = true;
									cuboid.points[i].movingId = touch.identifier;
									cuboid.points[i].prevX = ox;
									cuboid.points[i].prevY = oy;
								}
							}
						}
					}
				}
			},

			//=================================================================
			// キャンバスタッチ終了
			touchendCanvas1: function(e) { this.touchendCanvas(e, 1); },
			touchendCanvas2: function(e) { this.touchendCanvas(e, 2); },
			touchendCanvas: function(e, id) {
				var x, y;
				var touches = e.originalEvent.touches;  // 現在のタッチ情報
				var cuboid = editModel.cuboids[id-1];
				
				// フラグ
				var touched;
				
				// 実行
				for(var i=0; i<cuboid.points.length; i++) {
					touched = false;
					for(var j=0; j<touches.length; j++) {
						if(cuboid.points[i].moving && cuboid.points[i].movingId === touches[j].identifier) {
							touched = true;
							break;
						}
					}
					if(!touched) {
						cuboid.points[i].moving = false;
						cuboid.points[i].movingId = -1;
					}
				}
			},

			//=================================================================
			// キャンバス上でスライド操作
			touchmoveCanvas1: function(e) { return this.touchmoveCanvas(e, 1); },
			touchmoveCanvas2: function(e) { return this.touchmoveCanvas(e, 2); },
			touchmoveCanvas: function(e, id) {
				var touch, x, y;
				var touches = e.originalEvent.touches;
				var $canvas = $('#canvas' + id);
				var cuboid = editModel.cuboids[id-1];
				
				// フラグ初期化
				var pointsMoved = false;
				
				// 実行
				if(touches.length) {
					for(var j=0; j<touches.length; j++) {
						touch = touches[j];
						ox = touch.clientX - $canvas.offset().left + $(document).scrollLeft();
						oy = touch.clientY - $canvas.offset().top + $(document).scrollTop();
						for(var i=0; i<cuboid.points.length; i++) {
							if(cuboid.points[i].moving && cuboid.points[i].movingId === touch.identifier) {
								cuboid.points[i].x += ox - cuboid.points[i].prevX;
								cuboid.points[i].y += oy - cuboid.points[i].prevY;
								cuboid.points[i].prevX = ox;
								cuboid.points[i].prevY = oy;
								if(!pointsMoved) {
									pointsMoved = true;
								}
							}
						}
					}
					
					// 頂点が動いていた場合は再描画
					if(pointsMoved) {
						editController.redraw(id);
						return false;  // ブラウザスクロールさせない
					}
				}
				// どの頂点も操作対象でない場合は、ブラウザスクロールさせる
				return true;
			},

			//=================================================================
			// キャンバス上でのタッチがシステムによってキャンセルされた
			touchcancelCanvas1: function(e) { this.touchcancelCanvas(e, 1); },
			touchcancelCanvas2: function(e) { this.touchcancelCanvas(e, 2); },
			touchcancelCanvas: function(e, id) {
				var cuboid = editModel.cuboids[id-1];
				for(var i=0; i<cuboid.points.length; i++) {
					cuboid.points[i].moving = false;
					cuboid.points[i].movingId = -1;
				}
			},
			//=================================================================
			// PC対応。マウス操作を拾ってタッチイベントに投げる。
			
			// マウスダウン
			mousedownCanvas1: function(e) { this.mousedownCanvas(e, 1); },
			mousedownCanvas2: function(e) { this.mousedownCanvas(e, 2); },
			mousedownCanvas: function(e, id) {
				// イベントデータ生成
				e.originalEvent.touches = new Array(1);
				e.originalEvent.touches[0] = {
					identifier: 0  // 0 で決め打ち
					, clientX: e.originalEvent.clientX
					, clientY: e.originalEvent.clientY
				};
				
				// デリゲート
				this.touchstartCanvas(e, id);
			},

			// マウスアップ
			mouseupCanvas1: function(e) { this.mouseupCanvas(e, 1); },
			mouseupCanvas2: function(e) { this.mouseupCanvas(e, 2); },
			mouseupCanvas: function(e, id) {
				// イベントデータ生成
				e.originalEvent.touches = new Array(1);
				e.originalEvent.touches[0] = {
					identifier: 0  // 0 で決め打ち
					, clientX: e.originalEvent.clientX
					, clientY: e.originalEvent.clientY
				};
				
				// デリゲート
				this.touchcancelCanvas(e, id);  // touchend ではダメ
			},

			// マウス移動
			mousemoveCanvas1: function(e) { this.mousemoveCanvas(e, 1); },
			mousemoveCanvas2: function(e) { this.mousemoveCanvas(e, 2); },
			mousemoveCanvas: function(e, id) {
				var cuboid = editModel.cuboids[id-1];
				var moving = false;
				if(cuboid.points) {
					// 1点でも稼働中の点があるか？ない場合は反応しない。
					for(var i=0; i<cuboid.points.length; i++) {
						if(cuboid.points[i].moving) {
							moving = true;
							break;
						}
					}
					if(moving) {
						// イベントデータ生成
						e.originalEvent.touches = new Array(1);
						e.originalEvent.touches[0] = {
							identifier: 0  // 0 で決め打ち
							, clientX: e.originalEvent.clientX
							, clientY: e.originalEvent.clientY
						};
						
						// デリゲート
						this.touchmoveCanvas(e, id);
					}
				}
			},

			//=================================================================
			// 3Dデータを生成
			clickCreate3d: function() {
				var $preview3d = $('#preview3d');
				
				// 編集モードの場合のみ cuboid3d 生成
				if(editModel.mode === 'edit') {
					//------------------------
					// 既に生成済みの場合は生成されたものを削除してから生成する
					if(editModel.cuboid3d) {
						// モデルへの参照を削除
						//editModel.cuboid3d = null;
	
						// three.js で管理している描画領域を初期化
						//if(editModel.scene) {
						//	editModel.scene = null;
						//}
						$preview3d.html('');
						var s = '';
						for(var i=0; i<6; i++) {
							s += '<canvas id="texture' + i + '" style="display:none;" />';
						}
						$preview3d.html(s);
					}
					// min-height を設定
					$preview3d.css('min-height', $preview3d.width() + 'px');
	
					//------------------------
					// 3D データを editModel.cuboid3d に生成
					editUtil.generateCuboid3d();
				}
				
				// three.js の世界に cuboid3d をぶっこむ。
				threeUtil.resetPreview3d();
				
				// サイズ調整コントロールを表示
				$('#adjustSize').css('display', '');
				
				//------------------------
				// 再生成対応
				// サイズ調整コントロール
				$('#cuboidWidth').val(editModel.cuboid3d.width);
				$('#cuboidHeight').val(editModel.cuboid3d.height);
				$('#cuboidDepth').val(editModel.cuboid3d.depth);
				
				//------------------------
				// 次の入力項目を有効化
				$('#saveLocal').removeAttr("disabled");
				$('#upload').removeAttr("disabled");
				
			},
			
			//=================================================================
			// 3D タッチ
			touchstartPreview3d: function(e) {
				var touch, ox, oy;
				var touches = e.originalEvent.touches;  // 現在のタッチ情報
				var $preview3d = $('#preview3d');
				var cuboid3d = editModel.cuboid3d;
				
				// 実行
				if(touches.length) {
					touch = touches[0];
					ox = touch.clientX - $preview3d.offset().left + $(document).scrollLeft();;
					oy = touch.clientY - $preview3d.offset().top + $(document).scrollTop();
					
					// 状態保持
					cuboid3d.moving = true;
					cuboid3d.prevX = ox;
					cuboid3d.prevY = oy;
				}
			},
			
			// タッチ終了
			touchendPreview3d: function(e) {
				editModel.cuboid3d.moving = false;
			},
			
			// スライド
			touchmovePreview3d: function(e) {
				var touch, x, y, dx, dy, ox, oy;
				var touches = e.originalEvent.touches;
				var $preview3d = $('#preview3d');
				var cuboid3d = editModel.cuboid3d;
				var p = editModel.camera.position;
				
				// フラグ初期化
				var moved = false;
				
				// 実行
				if(touches.length && cuboid3d.moving) {
					touch = touches[0];
					ox = touch.clientX - $preview3d.offset().left + $(document).scrollLeft();
					oy = touch.clientY - $preview3d.offset().top + $(document).scrollTop();
					
					// 移動距離算出
					dx = ox - cuboid3d.prevX;
					dy = oy - cuboid3d.prevY;
					cuboid3d.prevX = ox;
					cuboid3d.prevY = oy;
					
					// カメラ位置操作
					threeUtil.moveCamera(dx, dy);
				}
				
				// いかなる場合もブラウザスクロールさせない
				return false;
			},
			
			// タッチがシステムによってキャンセルされた
			touchcancelPreview3d: function(e) {
				editModel.cuboid3d.moving = false;
			},
			
			//=================================================================
			// 3D マウス操作 (PC対応)
			
			// マウスダウン
			mousedownPreview3d: function(e) {
				// イベントデータ生成
				e.originalEvent.touches = new Array(1);
				e.originalEvent.touches[0] = {
					identifier: 0  // 0 で決め打ち
					, clientX: e.originalEvent.clientX
					, clientY: e.originalEvent.clientY
				};
				
				// デリゲート
				this.touchstartPreview3d(e);
			},
			
			// マウスアップ
			mouseupPreview3d: function(e) {
				// イベントデータ生成
				e.originalEvent.touches = new Array(1);
				e.originalEvent.touches[0] = {
					identifier: 0  // 0 で決め打ち
					, clientX: e.originalEvent.clientX
					, clientY: e.originalEvent.clientY
				};
				
				// デリゲート
				this.touchcancelPreview3d(e);
			},
			
			// マウス移動
			mousemovePreview3d: function(e) {
				// イベントデータ生成
				e.originalEvent.touches = new Array(1);
				e.originalEvent.touches[0] = {
					identifier: 0  // 0 で決め打ち
					, clientX: e.originalEvent.clientX
					, clientY: e.originalEvent.clientY
				};
				
				// デリゲート
				this.touchmovePreview3d(e);
			},
			
			//=================================================================
			// 3D ジェスチャー
			
			// ジェスチャー開始
			gesturestartPreview3d: function(e) {
				editModel.scaling = true;
				editModel.firstRadius = editModel.radius;
			},
			
			// ジェスチャー中に状態が変化
			gesturechangePreview3d: function(e) {
				// 拡大・縮小
				if(editModel.scaling) {
					editModel.radius = editModel.firstRadius / e.originalEvent.scale;
				}
				
				// レンダリング
				if(!editModel.autoRotation) {  // 自動回転が無効の時はレンダリング
					editModel.renderer.render(editModel.scene, editModel.camera);
				}
				
				// ブラウザに通知しない
				return false;
			},
			
			// ジェスチャー終了
			gestureendPreview3d: function(e) {
				editModel.scaling = false;
			},
			
			//=================================================================
			// 3D 自動回転
			changeAutoRotation: function(e) {
				if($('#autoRotation').val() === "true") {
					editModel.autoRotation = true;
				} else {
					editModel.autoRotation = false;
				}
			},
			
			//=================================================================
			// 3D サイズ調整
			changeCuboidWidth: function(e) {
				threeUtil.setWidth($('#cuboidWidth').val());
				threeUtil.resetCuboid3dPlatePosition();
			},
			changeCuboidHeight: function(e) {
				threeUtil.setHeight($('#cuboidHeight').val());
				threeUtil.resetCuboid3dPlatePosition();
			},
			changeCuboidDepth: function(e) {
				threeUtil.setDepth($('#cuboidDepth').val());
				threeUtil.resetCuboid3dPlatePosition();
			},
			
			//=================================================================
			// 保存
			clickSaveLocal: function() {
				if(window.localStorage) {
					// 情報を取得
					var rawData = localStorage.getItem(editModel.itemId);
					var cubox = null;
					if(rawData) {
						cubox = JSON.parse(rawData);
					} else {
						cubox = { cuboids: new Array() };
					}
					
					// データ生成
					var data = dataUtil.createCuboidJSON();
					
					// ローカルストレージに保存
					if(editModel.localCuboidId < 0) {
						// 初めての保存
						cubox.cuboids.push(data);
						editModel.localCuboidId = cubox.cuboids.length - 1;
						
					} else {
						// 上書き保存
						cubox.cuboids[editModel.localCuboidId] = data;
						
					}
					try {
						console.log(JSON.stringify(cubox).length);
						localStorage.setItem(editModel.itemId, JSON.stringify(cubox));
					} catch (e) {
						alert("ローカルストレージに保存できません。\n" + e);
						return;
					}
					
					// プレビューに出力
					$('#jsonPreview').val(JSON.stringify(data));
					
					// 保存成功メッセージ
					console.log(cubox);
					alert("ローカルストレージに保存しました。");
					
				} else {
					alert("お使いのブラウザはローカルストレージに対応していないため保存できません。");
				}
			},
			
			// アップロード
			clickUpload: function() {
				// メッセージ表示
				alert("現在アップロードはできません。");
				
				// データ生成
				//var data = dataUtil.createCuboidJSON();
				
				// アップロード
				
			}
			
		});
		return EditView;
	}
);
